/* if PRIORDs are not defined msg and sem queues are FIFOs otherwise the      */
/* insertion is in priority order and receiving tasks have their priority     */
/* boosted to that of the sending task, if it is higher than their base one   */

#define NSEM_PRIORD
#define MSG_PRIORD

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/malloc.h>

#include <asm/system.h>
#include <asm/rt_irq.h>

#include <linux/kernel.h>
#include <linux/timex.h>
#include <linux/cons.h>

#include <linux/rt_sched.h>

#include <asm/segment.h>

extern int rt_request_timer(void (*rt_timer_handler)(void), int rt_tick_period);

extern void rt_free_timer(void);

extern RTIME rt_linux_clock;

extern RTIME rt_base_time;

#define RT_TIME rt_base_time

#define RT_TICKS_PER_SEC 1193180LL

#define RT_TICKS_PER_SEC_H (RT_TICKS_PER_SEC/2 - 1)

#define RT_TIME_END 0x7fffFfffFfffFfffLL

#define RT_TASK_MAGIC 0x754d2774

#define RT_SEM_MAGIC 0xaabcdeff

#define SEM_ERR (0xFfff)

#define MSG_ERR ((RT_TASK *)0xFfff)

#define NOTHING ((void *)0)

#define SOMETHING ((void *)1)

#define __STR(x) #x
#define STR(x) __STR(x)

#define rt_switch_to(tsk) \
	__asm__ __volatile__( \
	"pushl %%eax\n\t" \
	"pushl %%ebp\n\t" \
	"pushl %%edi\n\t" \
	"pushl %%esi\n\t" \
	"pushl %%edx\n\t" \
	"pushl %%ecx\n\t" \
	"pushl %%ebx\n\t" \
	"movl "SYMBOL_NAME_STR(rt_current)", %%edx\n\t" \
	"pushl $1f\n\t" \
	"movl %%esp, (%%edx)\n\t" \
	"movl (%%ecx), %%esp\n\t" \
	"movl %%ecx, "SYMBOL_NAME_STR(rt_current)"\n\t" \
	"ret\n\t" \
"1:	popl %%ebx\n\t \
	popl %%ecx\n\t \
	popl %%edx\n\t \
	popl %%esi\n\t \
	popl %%edi\n\t \
	popl %%ebp\n\t \
	popl %%eax\n\t" \
	: /* no output */ \
	: "c" (tsk) \
	:"cx");

static RT_TASK *rt_current;

static RT_TASK *fpu_task;

static RT_TASK rt_linux_task;

static int linux_irq_state;

static int rt_latch;

static int rt_latch_h;

static RTIME rt_time_h;

/* ++++++++++++++++++++++++++++++++ TASKS ++++++++++++++++++++++++++++++++++ */

static void rt_thread_return(void)
{
	extern int rt_task_delete(RT_TASK *task);
	rt_task_delete(rt_current);
}


extern int rt_startup(void);
__asm__(".align 4; .globl "SYMBOL_NAME_STR(rt_startup)"\n\t"
		SYMBOL_NAME_STR(rt_startup)": iret");


int rt_task_init(RT_TASK *task, void (*rt_thread)(int), int data,
			int stack_size, int priority, int uses_fpu,
			void(*signal)(void))
{
	int *st;
	int flags;

	if (task->magic == RT_TASK_MAGIC) {
		return -EINVAL;
	}

	st = (int *) kmalloc(stack_size, GFP_KERNEL);	
	if (!st) {
		return -ENOMEM;
	}

	task->stack = st + stack_size / sizeof(int);	
	task->uses_fpu = uses_fpu;
	task->stack_bottom = st;
	task->magic = RT_TASK_MAGIC; 
	task->state = (SUSPENDED | READY);
	task->priority = task->base_priority = priority;
	task->period = 0;
	task->resume_time = RT_TIME_END;
	task->queue.prev = &(task->queue);	
	task->queue.next = &(task->queue);	
	task->queue.task = task;
	task->msg_queue.prev = &(task->msg_queue);	
	task->msg_queue.next = &(task->msg_queue);	
	task->msg_queue.task = task;	
	task->msg = 0;	
	task->blocked_on.task = NOTHING;	
	task->signal = signal;

	*--(task->stack) = data;
	*--(task->stack) = (int) rt_thread_return; /* in case a task returns */
	*--(task->stack) = 0x200;
	*--(task->stack) = KERNEL_CS;
	*--(task->stack) = (int) rt_thread;
	*--(task->stack) = (int) rt_startup;

	r_save_flags(flags);
	r_cli();
	if (uses_fpu) {
		__asm__("fsave %0" : "=m" (fpu_task->fpu_reg));
		__asm__("finit");
		__asm__("fsave %0" : "=m" (task->fpu_reg));
		__asm__("frstor %0" : "=m" (fpu_task->fpu_reg));
	}
	rt_linux_task.prev->next = task;
	task->prev = rt_linux_task.prev;
	task->next = 0;
	rt_linux_task.prev = task;
	r_restore_flags(flags);
	return 0;
}


static void rt_schedule(void)
{
	RT_TASK *task;
	RT_TASK *new_task;
	int prio;

	task = new_task = &rt_linux_task;
	prio = RT_LOWEST_PRIORITY;
	while ((task = task->next)) {
		if (task->state == READY && task->priority < prio) {
			new_task = task;
			prio = task->priority;
		}
	}

	if (new_task != rt_current) {
		if (new_task == &rt_linux_task) {
			SFIF = linux_irq_state;
		} else if (rt_current == &rt_linux_task) {
			linux_irq_state = SFIF;
			SFIF = 0;
		}
		if (new_task->uses_fpu && new_task != fpu_task) {
			__asm__("fsave %0" : "=m" (fpu_task->fpu_reg));
			fpu_task = new_task;
			__asm__("frstor %0" : "=m" (fpu_task->fpu_reg));
		}
		rt_switch_to(new_task);
		if (rt_current->signal) {
			(*rt_current->signal)();
		}
	}
}


int rt_task_delete(RT_TASK *task)
{
	int flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}

	r_save_flags(flags);
	r_cli();
	if (task->state & SEMAPHORE) {
		(task->blocked_on.sem)->count++;
		(task->queue.prev)->next = task->queue.next;
		(task->queue.next)->prev = task->queue.prev;
	} else if (task->state & (SEND | RPC)) {
		(task->queue.prev)->next = task->queue.next;
		(task->queue.next)->prev = task->queue.prev;
	} else {
		q = &(task->msg_queue);
		while ((q = q->next) != &(task->msg_queue)) {
			(q->task)->state &= ~(SEND | RPC | DELAYED);
		}	
	}
	task->magic = 0;
	if (!task->next) {
		rt_linux_task.prev = task->prev;
	}
	(task->prev)->next = task->next;
	(task->next)->prev = task->prev;
	kfree(task->stack_bottom);
	if (task == rt_current) {
		rt_schedule();
	}
	r_restore_flags(flags);
	return 0;
}


void rt_save_init_fpu()
{
	__asm__("fsave %0" : "=m" (fpu_task->fpu_reg));
	__asm__("finit");
}


void rt_restore_fpu(void)
{
	__asm__("frstor %0" : "=m" (fpu_task->fpu_reg));
}


static void rt_timer_handler(void)
{
	RT_TASK *task;
	RT_TASK *new_task;
	int prio;

	rt_time_h = (RT_TIME += rt_latch) + rt_latch_h;
	if (!(SFREQ & 1) && (RT_TIME >= rt_linux_clock)) {
		rt_linux_clock += LATCH;
		SFREQ |= 1;
	}

	task = new_task = &rt_linux_task;
	prio = RT_LOWEST_PRIORITY;
	while ((task = task->next)) {
		if ((task->state & DELAYED) && task->resume_time <= rt_time_h) {
			if (task->state & (SEMAPHORE | SEND | RPC)) {
				(task->queue.prev)->next = task->queue.next;
				(task->queue.next)->prev = task->queue.prev;
			}
			task->state &= ~(DELAYED | SEMAPHORE | RECEIVE |
					 SEND    | RPC       | RETURN);
		}
		if (task->state == READY && task->priority < prio) {
			new_task = task;
			prio = task->priority;
		}
	}

	if (new_task != rt_current) {
		if (new_task == &rt_linux_task) {
			SFIF = linux_irq_state;
		} else if (rt_current == &rt_linux_task) {
			linux_irq_state = SFIF;
			SFIF = 0;
		}
		if (new_task->uses_fpu && new_task != fpu_task) {
			__asm__("fsave %0" : "=m" (fpu_task->fpu_reg));
			fpu_task = new_task;
			__asm__("frstor %0" : "=m" (fpu_task->fpu_reg));
		}
		rt_switch_to(new_task);
		if (rt_current->signal) {
			(*rt_current->signal)();
		}
	}
}


RTIME start_rt_timer(int period)
{
	rt_latch = rt_request_timer(rt_timer_handler,
					period < LATCH ? period : LATCH);
	rt_latch_h = (rt_latch + 1)>>1;
	return rt_latch;
}


void stop_rt_timer(void)
{
	rt_free_timer();
}


static void signal_linux_timer(void)
{
/* recover all pending linux interrupts and (is it possible?) lost timings */
	while (SFIF & SFREQ) {
		sti();
		r_cli();
		if (!(SFREQ & 1) && (RT_TIME >= rt_linux_clock)) {
			rt_linux_clock += LATCH;
			SFREQ |= 1;
		}
		r_sti();
	}
}


RT_TASK *rt_whoami(void)
{
	return rt_current;
}


int rt_task_signal_handler(RT_TASK *task, void (*handler)(void))
{
	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	task->signal = handler;
	return 0;
}


int rt_task_use_fpu(RT_TASK *task, int use_fpu_flag)
{
	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	task->uses_fpu = use_fpu_flag != 0 ? 1 : 0;
	return 0;
}


int init_module(void)
{
	rt_linux_task.uses_fpu = 1;
	rt_linux_task.magic = 0;
	rt_linux_task.state = READY;
	rt_linux_task.priority = rt_linux_task.base_priority = RT_LOWEST_PRIORITY;
	rt_linux_task.signal = signal_linux_timer;
	rt_linux_task.prev = &rt_linux_task;
	rt_linux_task.next = 0;
	rt_current = &rt_linux_task;
	fpu_task = &rt_linux_task;
	printk("\n***** STARTING THE REAL TIME SCHEDULER WITH LINUX FREQUENCY (%d HZ) *****\n\n",(int)(RT_TICKS_PER_SEC/LATCH));
	start_rt_timer((int)LATCH);
	return 0;
}


void cleanup_module(void)
{
	stop_rt_timer();
	while (rt_linux_task.next) {
		rt_task_delete(rt_linux_task.next);
	}
	printk("\n***** THE REAL TIME SCHEDULER HAS BEEN REMOVED *****\n\n");
}

/* ++++++++++++++++++++++++++ TIME CONVERSIONS +++++++++++++++++++++++++++++ */

union time_union {RTIME rt_time; unsigned int time[2];};

typedef union time_union TIME_UNION;


static inline RTIME c2n(TIME_UNION timercounts)
{
	TIME_UNION high,low;
	unsigned int q;

	high.rt_time = timercounts.time[1]*1000000000LL;
	low.rt_time = timercounts.time[0]*1000000000LL;
	high.rt_time += low.time[1];
	q = high.rt_time/RT_TICKS_PER_SEC;
	low.time[1] = high.rt_time - q*RT_TICKS_PER_SEC;
	low.rt_time = (low.rt_time + RT_TICKS_PER_SEC_H)/RT_TICKS_PER_SEC;
	low.time[1] += q;
	return low.rt_time;
}


static inline RTIME n2c(TIME_UNION nanosecs)
{
	TIME_UNION high,low;
	unsigned int q;

	high.rt_time = nanosecs.time[1]*RT_TICKS_PER_SEC;
	low.rt_time = nanosecs.time[0]*RT_TICKS_PER_SEC;
	high.rt_time += low.time[1];
	q = high.rt_time/1000000000LL;
	low.time[1] = high.rt_time - q*1000000000LL;
	low.rt_time = (low.rt_time + 499999999LL)/1000000000LL;
	low.time[1] += q;
	return low.rt_time;
}


RTIME count2nano(RTIME timercounts)
{
	return c2n((TIME_UNION)timercounts);
}


RTIME nano2count(RTIME nanosecs)
{
	return 	n2c((TIME_UNION)nanosecs);
}

/* +++++++++++++++++++++++++++++++ TIMINGS ++++++++++++++++++++++++++++++++++ */

RTIME rt_get_time(void)
{
	int flags;
	RTIME now;

	r_save_flags(flags);
	r_cli();
	now = RT_TIME;
	r_restore_flags(flags);
	return now;
}


void rt_task_yield(void)
{
	int flags;

	r_save_flags(flags);
	r_cli();
	if (rt_current->next) {
		(rt_current->prev)->next = rt_current->next;
		(rt_current->next)->prev = rt_current->prev;
		rt_linux_task.prev->next = rt_current;
		rt_current->prev = rt_linux_task.prev;
		rt_current->next = 0;
		rt_linux_task.prev = rt_current;
		rt_schedule();
	}
	r_restore_flags(flags);
}


int rt_task_suspend(RT_TASK *task)
{
	int flags;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}

	r_save_flags(flags);
	r_cli();
	task->state |= SUSPENDED;
	if (task == rt_current) {
		rt_schedule();
	}
	r_restore_flags(flags);
	return 0;
}


int rt_task_resume(RT_TASK *task)
{
	int flags;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}

	r_save_flags(flags);
	r_cli();
	if (((task->state &= ~SUSPENDED) & ~DELAYED) == READY) {
		rt_schedule();
	}
	r_restore_flags(flags);
	return 0;
}


int rt_task_make_periodic(RT_TASK *task, RTIME start_time, RTIME period)
{
	long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	r_save_flags(flags);
	r_cli();
	task->resume_time = start_time;
	task->period = period;
	task->state = (task->state &~SUSPENDED) | DELAYED;
	rt_schedule();
	r_restore_flags(flags);
	return 0;
}


void rt_task_wait_period(void)
{
	long flags;

	r_save_flags(flags);
	r_cli();
	if ((rt_current->resume_time += rt_current->period) > rt_time_h) {
		rt_current->state |= DELAYED;
		rt_schedule();
	}
	r_restore_flags(flags);
}


RTIME next_period(void)
{
	return rt_current->resume_time + rt_current->period;
}


void rt_sleep(RTIME delay)
{
	int flags;
	r_save_flags(flags);
	r_cli();
	if ((rt_current->resume_time = RT_TIME + delay) > rt_time_h) {
		rt_current->state |= DELAYED;
		rt_schedule();
	}
	r_restore_flags(flags);
}


void rt_sleep_until(RTIME time)
{
	int flags;
	r_save_flags(flags);
	r_cli();
	if ((rt_current->resume_time = time) > rt_time_h) {
		rt_current->state |= DELAYED;
		rt_schedule();
	}
	r_restore_flags(flags);
}

/* +++++++++++++++++++++++++++++ SEMAPHORES ++++++++++++++++++++++++++++++++ */

void rt_sem_init(SEM *sem, int value)
{
	sem->count = value;
	sem->magic = RT_SEM_MAGIC;
	sem->queue.prev = &(sem->queue);
	sem->queue.next = &(sem->queue);
	sem->queue.task = 0;
}


int rt_sem_delete(SEM *sem)
{
	int flags;
	QUEUE *q;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	r_save_flags(flags);
	r_cli();
	sem->magic = 0;
	q = &(sem->queue);
	while ((q = q->next) != &(sem->queue)) {
		(q->task)->state &= ~(SEMAPHORE | DELAYED);
	}
	r_restore_flags(flags);
	return 0;
}


int rt_sem_signal(SEM *sem)
{
	int flags;
	RT_TASK *task;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	r_save_flags(flags);
	r_cli();
	if ((sem->count)++ < 0) {
		if ((task = (sem->queue.next)->task)) {
			sem->queue.next = task->queue.next;
			(task->queue.next)->prev = &(sem->queue);
			task->blocked_on.sem = NOTHING;
			if ((task->state &= ~(SEMAPHORE | DELAYED)) == READY) {
				rt_schedule();
			}
		}
	}
	r_restore_flags(flags);
	return 0;
}


int rt_sem_wait(SEM *sem)
{
	int flags;
	int count;
	QUEUE *q;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	r_save_flags(flags);
	r_cli();
	count = sem->count;
	if (--(sem->count) < 0) {
		rt_current->state |= SEMAPHORE;
		rt_current->blocked_on.sem = sem;
		q = &(sem->queue);
#ifdef SEM_PRIORD
		while ((q = q->next) != &(sem->queue) &&
			 (q->task)->priority <= rt_current->priority);
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_schedule();
		if (rt_current->blocked_on.sem || sem->magic != RT_SEM_MAGIC) {
			count = SEM_ERR;
		} else { 
			count = sem->count;
		}
	}
	r_restore_flags(flags);
	return count;
}


int rt_sem_wait_if(SEM *sem)
{
	int count;
	int flags;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	r_save_flags(flags);
	r_cli();
	if ((count = sem->count) > 0) {
		sem->count--;
	}
	r_restore_flags(flags);
	return count;
}


int rt_sem_wait_until(SEM *sem, RTIME time)
{
	int count;
	int flags;
	QUEUE *q;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	r_save_flags(flags);
	r_cli();
	count = sem->count;
	if (--(sem->count) < 0) {
		rt_current->blocked_on.sem = sem;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->state |= (SEMAPHORE | DELAYED);
			q = &(sem->queue);
#ifdef SEM_PRIORD
			while ((q = q->next) != &(sem->queue) &&
				 (q->task)->priority <= rt_current->priority);
#endif
			rt_current->queue.prev = q->prev;
			rt_current->queue.next = q;
			(q->prev)->next = &(rt_current->queue);
			q->prev = &(rt_current->queue);
			rt_schedule();
		}
		if (sem->magic != RT_SEM_MAGIC) {
			count = SEM_ERR;
		} else {
			if (rt_current->blocked_on.sem) {
				rt_current->blocked_on.sem = NOTHING;
				count = sem->count;
				sem->count++;
			} else {
				count = sem->count;
			}
		}
	}
	r_restore_flags(flags);
	return count;
}


int rt_sem_wait_timed(SEM *sem, RTIME delay)
{
	return rt_sem_wait_until(sem, RT_TIME + delay);
}

/* ++++++++++++++++++++++++++++++ MESSAGES +++++++++++++++++++++++++++++++++ */

/* ++++++++++++++++++++++++++++++++ SEND +++++++++++++++++++++++++++++++++++ */

RT_TASK *rt_send(RT_TASK *task, unsigned int msg)
{
	int flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	r_save_flags(flags);
	r_cli();
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = msg;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		if ((task->state &= ~(RECEIVE | DELAYED)) == READY) {
			rt_schedule();
		}
	} else {
		rt_current->msg = msg;
		rt_current->msg_queue.task = task;
		q = &(task->msg_queue);
#ifdef MSG_PRIORD
		while ((q = q->next) != &(task->msg_queue) &&
			 (q->task)->priority <= rt_current->priority);
		if (task->priority > rt_current->priority) {
			task->priority = rt_current->priority;
		}
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_current->state |= SEND;
		rt_schedule();
	}
	if (rt_current->msg_queue.task != rt_current) {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	r_restore_flags(flags);
	return task;
}


RT_TASK *rt_send_if(RT_TASK *task, unsigned int msg)
{
	int flags;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	r_save_flags(flags);
	r_cli();
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = msg;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		if ((task->state &= ~(RECEIVE | DELAYED)) == READY) {
			rt_schedule();
		}
		if (rt_current->msg_queue.task != rt_current) {
			rt_current->msg_queue.task = rt_current;
			task = (RT_TASK *)0;
		}
	} else {
		task = (RT_TASK *)0;
	}
	r_restore_flags(flags);
	return task;
}


RT_TASK *rt_send_until(RT_TASK *task, unsigned int msg, RTIME time)
{
	int flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	r_save_flags(flags);
	r_cli();
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = msg;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		if ((task->state &= ~(RECEIVE | DELAYED)) == READY) {
			rt_schedule();
		}
	} else {
		rt_current->msg_queue.task = task;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->msg = msg;
			q = &(task->msg_queue);
#ifdef MSG_PRIORD
			while ((q = q->next) != &(task->msg_queue) &&
				 (q->task)->priority <= rt_current->priority);
			if (task->priority > rt_current->priority) {
				task->priority = rt_current->priority;
			}
#endif
			rt_current->queue.prev = q->prev;
			rt_current->queue.next = q;
			(q->prev)->next = &(rt_current->queue);
			q->prev = &(rt_current->queue);
			rt_current->state |= (SEND | DELAYED);
			rt_schedule();
		}
	}
	if (rt_current->msg_queue.task != rt_current) {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	r_restore_flags(flags);
	return task;
}


RT_TASK *rt_send_timed(RT_TASK *task, unsigned int msg, RTIME delay)
{
	return rt_send_until(task, msg, RT_TIME + delay);
}

/* +++++++++++++++++++++++++++++++++ RPC +++++++++++++++++++++++++++++++++++ */

RT_TASK *rt_rpc(RT_TASK *task, unsigned int to_do, unsigned int *result)
{

	int flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	r_save_flags(flags);
	r_cli();
	if ((task->state & RECEIVE) &&
		(!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = to_do;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		task->state &= ~(RECEIVE | DELAYED);
		rt_current->state |= RETURN;
	} else {
		rt_current->msg = to_do;
		q = &(task->msg_queue);
#ifdef MSG_PRIORD
		while ((q = q->next) != &(task->msg_queue) &&
			 (q->task)->priority <= rt_current->priority);
		if (task->priority > rt_current->priority) {
			task->priority = rt_current->priority;
		}
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_current->state |= RPC;
	}
	rt_current->msg_queue.task = task;
	rt_schedule();
	if (rt_current->msg_queue.task == rt_current) {
		*result = rt_current->msg;
	} else {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	r_restore_flags(flags);
	return task;
}


RT_TASK *rt_rpc_if(RT_TASK *task, unsigned int to_do, unsigned int *result)
{
	int flags;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}
	r_save_flags(flags);
	r_cli();
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = to_do;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		task->state &= ~(RECEIVE | DELAYED);
		rt_current->state |= RETURN;
		rt_current->msg_queue.task = task;
		rt_schedule();
		if (rt_current->msg_queue.task == rt_current) {
			*result = rt_current->msg;
		} else {
			rt_current->msg_queue.task = rt_current;
			task = (RT_TASK *)0;
		}
	} else {
		task = (RT_TASK *)0;
	}
	r_restore_flags(flags);
	return task;
}


RT_TASK *rt_rpc_until(RT_TASK *task, unsigned int to_do, unsigned int *result, RTIME time)
{
	int flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	r_save_flags(flags);
	r_cli();
	if ((task->state & RECEIVE) &&
	    (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = to_do;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		task->state &= ~(RECEIVE | DELAYED);
		rt_current->state |= (RETURN | DELAYED);
	} else {
		if ((rt_current->resume_time = time) <= rt_time_h) {
			r_restore_flags(flags);
			return (RT_TASK *)0;
		}
		rt_current->msg = to_do;
		q = &(task->msg_queue);
#ifdef MSG_PRIORD
		while ((q = q->next) != &(task->msg_queue) &&
			 (q->task)->priority <= rt_current->priority);
		if (task->priority > rt_current->priority) {
			task->priority = rt_current->priority;
		}
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_current->state |= (RPC | DELAYED);
	}
	rt_current->msg_queue.task = task;
	rt_schedule();
	if (rt_current->msg_queue.task == rt_current) {
		*result = rt_current->msg;
	} else {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	r_restore_flags(flags);
	return task;
}


RT_TASK *rt_rpc_timed(RT_TASK *task, unsigned int to_do, unsigned int *result, RTIME delay)
{
	return rt_rpc_until(task, to_do, result, RT_TIME + delay);
}

/* ++++++++++++++++++++++++++++++ RPC_RETURN +++++++++++++++++++++++++++++++ */

int rt_isrpc(RT_TASK *task)
{
	return task->state & RETURN;
}


RT_TASK *rt_return(RT_TASK *task, unsigned int result)
{
	int flags;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	r_save_flags(flags);
	r_cli();
	if ((task->state & RETURN) && task->msg_queue.task == rt_current) {
		task->msg = result;
		task->msg_queue.task = task;
		if ((task->state &= ~(RETURN | DELAYED)) == READY) {
			rt_schedule();
		}
	} else {
		task = (RT_TASK *)0;
	}
	r_restore_flags(flags);
	return task;
}

/* +++++++++++++++++++++++++++++++ RECEIVE +++++++++++++++++++++++++++++++++ */

RT_TASK *rt_receive(RT_TASK *task, unsigned int *msg)
{
	int flags;

	if (task && task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	r_save_flags(flags);
	r_cli();

	if (!task) task = (rt_current->msg_queue.next)->task;
	if ((task->state & (SEND | RPC)) && task->msg_queue.task == rt_current) {
		(task->queue.next)->prev = task->queue.prev;
		(task->queue.prev)->next = task->queue.next;
		*msg = task->msg;
#ifdef MSG_PRIORD
		rt_current->priority = rt_current->base_priority > task->priority ?
			task->priority : rt_current->base_priority;
#endif
		if (task->state & SEND) {
			task->msg_queue.task = task;
			if ((task->state  &=  ~(SEND | DELAYED)) == READY) {
				rt_schedule();
			}
		} else if (task->state & RPC) {
			task->state = (task->state & ~(RPC | DELAYED)) | RETURN;
		}
	} else {
#ifdef MSG_PRIORD
		task->priority = rt_current->base_priority;
#endif
		rt_current->blocked_on.task = SOMETHING;
		rt_current->state |= RECEIVE;
		rt_current->msg_queue.task = (task != rt_current) ? task : (RT_TASK *)0;
		rt_schedule();
		*msg = rt_current->msg;
	}
	if (rt_current->blocked_on.task) {
		rt_current->blocked_on.task = NOTHING;
		task = (RT_TASK *)0;
	} else {
               	task = rt_current->msg_queue.task;
	}
	rt_current->msg_queue.task = rt_current;
	r_restore_flags(flags);
	return task;
}


RT_TASK *rt_receive_if(RT_TASK *task, unsigned int *msg)
{
	int flags;

	if (task && task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	r_save_flags(flags);
	r_cli();

	if (!task) task = (rt_current->msg_queue.next)->task;
	if ((task->state & (SEND | RPC)) && task->msg_queue.task == rt_current) {
		(task->queue.next)->prev = task->queue.prev;
		(task->queue.prev)->next = task->queue.next;
		*msg = task->msg;
#ifdef MSG_PRIORD
		rt_current->priority = rt_current->base_priority > task->priority ?
			task->priority : rt_current->base_priority;
#endif
		if (task->state & SEND) {
			task->msg_queue.task = task;
			if ((task->state  &=  ~(SEND | DELAYED)) == READY) {
				rt_schedule();
			}
		} else if (task->state & RPC) {
			task->state = (task->state & ~(RPC | DELAYED)) | RETURN;
		}
		if (rt_current->blocked_on.task) {
			rt_current->blocked_on.task = NOTHING;
			task = (RT_TASK *)0;
		} else {
        	       	task = rt_current->msg_queue.task;
		}
		rt_current->msg_queue.task = rt_current;
	} else {
		task = (RT_TASK *)0;
	}
	r_restore_flags(flags);
	return task;
}


RT_TASK *rt_receive_until(RT_TASK *task, unsigned int *msg, RTIME time)
{
	int flags;

	if (task && task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	r_save_flags(flags);
	r_cli();

	if (!task) task = (rt_current->msg_queue.next)->task;
	if ((task->state & (SEND | RPC)) && task->msg_queue.task == rt_current) {
		(task->queue.next)->prev = task->queue.prev;
		(task->queue.prev)->next = task->queue.next;
		*msg = task->msg;
#ifdef MSG_PRIORD
		rt_current->priority = rt_current->base_priority > task->priority ?
			task->priority : rt_current->base_priority;
#endif
		if (task->state & SEND) {
			task->msg_queue.task = task;
			if ((task->state  &=  ~(SEND | DELAYED)) == READY) {
				rt_schedule();
			}
		} else if (task->state & RPC) {
			task->state = (task->state & ~(RPC | DELAYED)) | RETURN;
		}
	} else {
#ifdef MSG_PRIORD
		task->priority = rt_current->base_priority;
#endif
		rt_current->blocked_on.task = SOMETHING;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->state |= (RECEIVE | DELAYED);
			rt_current->msg_queue.task = (task != rt_current) ? task : (RT_TASK *)0;
			rt_schedule();
			*msg = rt_current->msg;
		}
	}
	if (rt_current->blocked_on.task) {
		rt_current->blocked_on.task = NOTHING;
		task = (RT_TASK *)0;
	} else {
		task = rt_current->msg_queue.task;
	}
	rt_current->msg_queue.task = rt_current;
	r_restore_flags(flags);
	return task;
}


RT_TASK *rt_receive_timed(RT_TASK *task, unsigned int *msg, RTIME delay)
{
	return rt_receive_until(task, msg, RT_TIME + delay);
}
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
