#ifdef __RT__

#include <asm/io.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <asm/rt_time.h>
#include <asm/rt_irq.h>
#include <linux/sched.h>
#include <linux/timex.h>

RTIME rt_base_time;

RTIME rt_linux_clock;

static int rt_latch;

RTIME rt_get_time(void)
{
	int flags;
	int c0;
	RTIME now;

	r_save_flags(flags);
	r_cli();
	outb(0x00, 0x43);
	c0 = inb(0x40);
	c0 |= (inb(0x40) << 8);
	now = rt_base_time + ((rt_latch - c0) & 0xffff);
	r_restore_flags(flags);
	return now;
}


int rt_request_timer(void (*timer_handler)(void), int latch) 
{
	int flags;

	r_save_flags(flags);
	r_cli();
	outb_p(0x34, 0x43);
	outb_p(latch & 0xff, 0x40);
	outb(latch >> 8, 0x40);
	free_RTirq(0);
	request_RTirq(0, timer_handler);
	rt_latch = latch;
	r_restore_flags(flags);
	return rt_latch;
}


static void linux_timer_handler(void)
{
    	SFREQ |= 1;
	rt_base_time = rt_linux_clock;
	rt_linux_clock += LATCH;
}


void rt_free_timer(void)
{
	rt_request_timer(linux_timer_handler, LATCH);
}


extern void rt_fifo_init(void);

void rt_init(void) 
{
	rt_base_time = 0;
	rt_linux_clock = LATCH;
	rt_fifo_init();
	rt_free_timer( );
}

#endif /* __RT__ */